@@ -29,31 +29,22 @@ class WindowFunction(Enum):
     NTILE = 'ntile'
     ROW_NUMBER = 'row_number'
 
-    _NAVIGATION_FUNCTIONS = (
-        FIRST_VALUE,
-        LAST_VALUE,
-        NTH_VALUE,
-        LEAD,
-        LAG,
-    )
-
-    _NUMBERING_FUNCTIONS = (
+    _BQ_FUNCTIONS_THAT_DO_NOT_SUPPORT_WINDOW_FRAME_CLAUSE = (
         RANK,
         DENSE_RANK,
         PERCENT_RANK,
         CUME_DIST,
         NTILE,
         ROW_NUMBER,
-    )
-
-    _ALL_FUNCTIONS = (
-        *_NAVIGATION_FUNCTIONS,
-        *_NUMBERING_FUNCTIONS,
+        LAG,
+        LEAD,
     )
 
     def supports_window_frame_clause(self, dialect: Dialect) -> bool:
         if is_bigquery(dialect):
-            return self.value not in WindowFunction._NUMBERING_FUNCTIONS.value
+            return (
+                self.value not in WindowFunction._BQ_FUNCTIONS_THAT_DO_NOT_SUPPORT_WINDOW_FRAME_CLAUSE.value
+            )
 
         if is_postgres(dialect):
             return True
@@ -345,6 +336,7 @@ class Window(GroupBy):
     def __init__(self,
                  group_by_columns: List['Series'],
                  order_by: List[SortColumn],
+                 nulls_last: bool = False,
                  mode: WindowFrameMode = WindowFrameMode.RANGE,
                  start_boundary: Optional[WindowFrameBoundary] = WindowFrameBoundary.PRECEDING,
                  start_value: int = None,
@@ -412,6 +404,7 @@ class Window(GroupBy):
                                 f' AND {end_boundary.frame_clause(end_value)}'
 
         self._order_by = order_by
+        self._nulls_last = nulls_last
 
     @property
     def frame_clause(self) -> str:
@@ -439,7 +432,9 @@ class Window(GroupBy):
         :see: __init__()
         """
         return Window(group_by_columns=list(self._index.values()),
-                      order_by=self._order_by, mode=mode,
+                      order_by=self._order_by,
+                      nulls_last=self._nulls_last,
+                      mode=mode,
                       start_boundary=start_boundary, start_value=start_value,
                       end_boundary=end_boundary, end_value=end_value)
 
@@ -450,7 +445,11 @@ class Window(GroupBy):
         """
         if self._order_by:
             exprs = [sc.expression for sc in self._order_by]
-            fmtstr = [f"{{}} {'asc' if sc.asc else 'desc'}" for sc in self._order_by]
+            nulls_last_stmt = 'nulls last' if self._nulls_last else ''
+            fmtstr = [
+                f"{{}} {'asc' if sc.asc else 'desc'} {nulls_last_stmt}".strip()
+                for sc in self._order_by
+            ]
             return Expression.construct(f'order by {", ".join(fmtstr)}', *exprs)
         else:
             return Expression.construct('')
